/*******************************************************************************
 * Copyright (c) 2000, 2006 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.internal.theme.Theme;

import dwt.dwthelper.utils;

import dwt.DWT;
import dwt.graphics.Device;
import dwt.graphics.GC;
import dwt.graphics.GCData;
import dwt.graphics.Image;
import dwt.graphics.ImageData;
import dwt.graphics.Point;
import dwt.graphics.Rectangle;
import dwt.internal.Converter;
import dwt.internal.gtk.OS;

public class Theme {
    Device device;
    
    int /*long*/ shellHandle, fixedHandle, buttonHandle, arrowHandle,
        frameHandle, entryHandle, checkButtonHandle, radioButtonHandle, 
        notebookHandle, treeHandle, progressHandle, toolbarHandle,
        labelHandle, separatorHandle;
    
public Theme(Device device) {
    this.device = device;
    shellHandle = OS.gtk_window_new (OS.GTK_WINDOW_TOPLEVEL);
    fixedHandle = OS.gtk_fixed_new();
    buttonHandle = OS.gtk_button_new();
    arrowHandle = OS.gtk_arrow_new(OS.GTK_ARROW_DOWN, OS.GTK_SHADOW_NONE);
    checkButtonHandle = OS.gtk_check_button_new();
    frameHandle = OS.gtk_check_button_new();
    entryHandle = OS.gtk_entry_new();
    radioButtonHandle = OS.gtk_radio_button_new(0);
    notebookHandle = OS.gtk_notebook_new();
    progressHandle = OS.gtk_progress_bar_new();
    toolbarHandle = OS.gtk_toolbar_new();
    treeHandle = OS.gtk_tree_view_new_with_model(0);
    separatorHandle = OS.gtk_vseparator_new();
    labelHandle = OS.gtk_label_new(null);
    OS.gtk_container_add (fixedHandle, labelHandle);
    OS.gtk_container_add (fixedHandle, frameHandle);
    OS.gtk_container_add (fixedHandle, entryHandle);
    OS.gtk_container_add (fixedHandle, separatorHandle);
    OS.gtk_container_add (fixedHandle, arrowHandle);
    OS.gtk_container_add (fixedHandle, toolbarHandle);
    OS.gtk_container_add (fixedHandle, progressHandle);
    OS.gtk_container_add (fixedHandle, checkButtonHandle);
    OS.gtk_container_add (fixedHandle, radioButtonHandle);
    OS.gtk_container_add (fixedHandle, buttonHandle);
    OS.gtk_container_add (fixedHandle, treeHandle);
    OS.gtk_container_add (fixedHandle, notebookHandle);
    OS.gtk_container_add (shellHandle, fixedHandle);
    OS.gtk_widget_realize (separatorHandle);
    OS.gtk_widget_realize (labelHandle);
    OS.gtk_widget_realize (frameHandle);
    OS.gtk_widget_realize (entryHandle);
    OS.gtk_widget_realize (arrowHandle);
    OS.gtk_widget_realize (buttonHandle);
    OS.gtk_widget_realize (treeHandle);
    OS.gtk_widget_realize (notebookHandle);
    OS.gtk_widget_realize (checkButtonHandle);
    OS.gtk_widget_realize (radioButtonHandle);
    OS.gtk_widget_realize (progressHandle);
    OS.gtk_widget_realize (toolbarHandle);
    OS.gtk_widget_realize (shellHandle);
}

//TODO - share this code 
static int /*long*/ createPixbuf(Image image) {
    int [] w = new int [1], h = new int [1];
    OS.gdk_drawable_get_size (image.pixmap, w, h);
    int /*long*/ colormap = OS.gdk_colormap_get_system ();
    int /*long*/ pixbuf;
    bool hasMask = image.mask !is 0 && OS.gdk_drawable_get_depth (image.mask) is 1;
    if (hasMask) {
        pixbuf = OS.gdk_pixbuf_new (OS.GDK_COLORSPACE_RGB, true, 8, w [0], h [0]);
        if (pixbuf is 0) DWT.error (DWT.ERROR_NO_HANDLES);
        OS.gdk_pixbuf_get_from_drawable (pixbuf, image.pixmap, colormap, 0, 0, 0, 0, w [0], h [0]);
        int /*long*/ maskPixbuf = OS.gdk_pixbuf_new(OS.GDK_COLORSPACE_RGB, false, 8, w [0], h [0]);
        if (maskPixbuf is 0) DWT.error (DWT.ERROR_NO_HANDLES);
        OS.gdk_pixbuf_get_from_drawable(maskPixbuf, image.mask, 0, 0, 0, 0, 0, w [0], h [0]);
        int stride = OS.gdk_pixbuf_get_rowstride(pixbuf);
        int /*long*/ pixels = OS.gdk_pixbuf_get_pixels(pixbuf);
        byte[] line = new byte[stride];
        int maskStride = OS.gdk_pixbuf_get_rowstride(maskPixbuf);
        int /*long*/ maskPixels = OS.gdk_pixbuf_get_pixels(maskPixbuf);
        byte[] maskLine = new byte[maskStride];
        for (int y=0; y<h[0]; y++) {
            int /*long*/ offset = pixels + (y * stride);
            OS.memmove(line, offset, stride);
            int /*long*/ maskOffset = maskPixels + (y * maskStride);
            OS.memmove(maskLine, maskOffset, maskStride);
            for (int x=0; x<w[0]; x++) {
                if (maskLine[x * 3] is 0) {
                    line[x * 4 + 3] = 0;
                }
            }
            OS.memmove(offset, line, stride);
        }
        OS.g_object_unref(maskPixbuf);
    } else {
        ImageData data = image.getImageData ();
        bool hasAlpha = data.getTransparencyType () is DWT.TRANSPARENCY_ALPHA;
        pixbuf = OS.gdk_pixbuf_new (OS.GDK_COLORSPACE_RGB, hasAlpha, 8, w [0], h [0]);
        if (pixbuf is 0) DWT.error (DWT.ERROR_NO_HANDLES);
        OS.gdk_pixbuf_get_from_drawable (pixbuf, image.pixmap, colormap, 0, 0, 0, 0, w [0], h [0]);
        if (hasAlpha) {
            byte [] alpha = data.alphaData;
            int stride = OS.gdk_pixbuf_get_rowstride (pixbuf);
            int /*long*/ pixels = OS.gdk_pixbuf_get_pixels (pixbuf);
            byte [] line = new byte [stride];
            for (int y = 0; y < h [0]; y++) {
                int /*long*/ offset = pixels + (y * stride);
                OS.memmove (line, offset, stride);
                for (int x = 0; x < w [0]; x++) {
                    line [x*4+3] = alpha [y*w [0]+x];
                }
                OS.memmove (offset, line, stride);
            }
        }
    }
    return pixbuf;
}

void checkTheme() {
    if (isDisposed()) DWT.error(DWT.ERROR_GRAPHIC_DISPOSED);
}

public Rectangle computeTrim(GC gc, DrawData data) {
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    return data.computeTrim(this, gc);
}

public void dispose () {
    if (shellHandle is 0) return;
    OS.gtk_widget_destroy(shellHandle);
    shellHandle = fixedHandle = buttonHandle = arrowHandle =
    frameHandle = entryHandle = checkButtonHandle = radioButtonHandle = 
    notebookHandle = treeHandle = progressHandle = toolbarHandle = 
    labelHandle = separatorHandle = 0;
}
    
public void drawBackground(GC gc, Rectangle bounds, DrawData data) {
    checkTheme();
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    data.draw(this, gc, bounds);
}

public void drawFocus(GC gc, Rectangle bounds, DrawData data) {
    checkTheme();
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    gc.drawFocus(bounds.x, bounds.y, bounds.width, bounds.height);
}

public void drawImage(GC gc, Rectangle bounds, DrawData data, Image image, int flags) {
    checkTheme();
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (image is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    data.drawImage(this, image, gc, bounds);
}

public void drawText(GC gc, Rectangle bounds, DrawData data, String text, int flags) {
    checkTheme();
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (text is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    data.drawText(this, text, flags, gc, bounds);
}

public Rectangle getBounds(int part, Rectangle bounds, DrawData data) {
    checkTheme();
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    return data.getBounds(part, bounds);
}

public int getSelection(Point offset, Rectangle bounds, RangeDrawData data) {
    checkTheme();
    if (offset is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    return data.getSelection(offset, bounds);
}

public int hitBackground(Point position, Rectangle bounds, DrawData data) {
    checkTheme();
    if (position is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (bounds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    return data.hit(this, position, bounds);
}

public bool isDisposed() {
    return device is null;
}

public Rectangle measureText(GC gc, Rectangle bounds, DrawData data, String text, int flags) {
    checkTheme();
    if (gc is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (data is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (text is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed()) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    return data.measureText(this, text, flags, gc, bounds);
}

int getWidgetProperty(int /*long*/ handle, String name) {
    byte[] propertyName = Converter.wcsToMbcs(null, name, true);
    int[] result = new int[1];
    OS.gtk_widget_style_get(handle, propertyName, result, 0);
    return result[0];
}

void transferClipping(GC gc, int /*long*/ style) {
    GCData data = gc.getGCData();
    int /*long*/ clipRgn = data.clipRgn;
    int /*long*/ damageRgn = data.damageRgn;
    int /*long*/ clipping = clipRgn;
    if (damageRgn !is 0) {
        if (clipping !is 0) {
            clipping = OS.gdk_region_new();
            OS.gdk_region_union(clipping, clipRgn);
            OS.gdk_region_intersect(clipping, damageRgn);
        } else {
            clipping = damageRgn;
        }
    }
    int /*long*/ [] curGC = new int /*long*/ [1];
    for (int i = 0; i < 5; i++) {
        OS.gtk_style_get_fg_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_bg_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_light_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_dark_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_mid_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_text_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
        OS.gtk_style_get_text_aa_gc (style, i, curGC);
        if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
    }
    OS.gtk_style_get_black_gc (style, curGC);
    if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
    OS.gtk_style_get_white_gc (style, curGC);
    if (curGC[0] !is 0) OS.gdk_gc_set_clip_region (curGC[0], clipping);
    if (clipping !is clipRgn && clipping !is damageRgn) {
        OS.gdk_region_destroy(clipping);
    }
}
}
